import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;
import java.util.Vector;
import java.util.concurrent.ConcurrentHashMap;

import com.inertia.orientation.ImuFilter;
import com.inertia.orientation.OrientationFilter;
import com.inertia.orientation.OrientationFilters;
import com.inertia.promove.Command;
import com.inertia.promove.Config;
import com.inertia.promove.Config.GlobalSamplingConfig;
import com.inertia.promove.Config.Mask;
import com.inertia.promove.InertiaException;
import com.inertia.promove.ProMoveSamples.Sample;
import com.inertia.promove.ProtocolListener;
import com.inertia.promove.SerialProtocol;
import com.inertia.promove.SerialProtocol.Message;
import com.inertia.promove.Transformer;

import jssc.SerialPort;
import jssc.SerialPortException;

public class ReadSerialPort implements ProtocolListener {

    static final String SERIAL_PORT = "COM10";
    static final boolean enableConfigChangeExample = false;

    volatile SerialPort serialPort = null;
    volatile SerialProtocol serialProtocol = null;
    ConfigRequestThread crt = null;
    ConfigChangeThread cct = null;

    boolean abort = false;

    /**
     * Main method
     *
     * @param args
     *            first argument is optional port
     */
    public static void main(String[] args) {
        new ReadSerialPort(args);
    }

    /**
     * Constructor
     *
     * @param args
     *            first argument is optional port
     */
    ReadSerialPort(String[] args) {

        System.err.println("Using: " + SerialProtocol.LIB_VERSION_STR);

        serialPort = new SerialPort(args.length > 0 ? args[0] : SERIAL_PORT);

        try {
            serialPort.openPort();
            serialPort.setParams(460800, 8, 1, 0);

            serialProtocol = new SerialProtocol(this);

            // keep track of nodes and their last timestamps
            HashMap<Integer, Double> lastTimestamp = new HashMap<Integer, Double>();
            Vector<Integer> nodes = new Vector<Integer>();

            // start a thread to automatically request configurations
            crt = new ConfigRequestThread();
            crt.start();

            // start a thread to automatically change configurations
            if (enableConfigChangeExample) {
                cct = new ConfigChangeThread();
                cct.start();
            }

            System.out.print(Sample.getCsvLegend());

            while (!abort) {

                // Read available bytes from serial port.
                byte[] buffer = serialPort.readBytes();
                if (buffer == null) {
                    try {
                        Thread.sleep(5);
                    } catch (InterruptedException e) {
                    }
                    continue;
                }

                // Parse bytes and analyze messages
                Vector<Message> messages = serialProtocol.parseSerialMessages(buffer);
                for (Message m : messages) {

                    // received configurations
                    for (Config config : m.getConfigs()) {
                        // indicate configuration is received and reset orientation filter.
                        int nodeID = m.getNodeID();
                        if (nodeID > 0) {
                            crt.configReceived(nodeID);
                            OrientationFilters.updateFilter(serialProtocol.toString(), nodeID, null);
                        }
                        System.err.println("Node: " + nodeID + "\n" + config.toString());
                    }

                    // received commands
                    for (Command command : m.getCommands()) {
                        System.err.println(command.toString());
                    }

                    // received samples
                    for (Sample s : m.getSamples()) {
                        // make sure configuration of node is known
                        if (!nodes.contains(s.nodeID)) {
                            crt.requestConfig(s.nodeID, Mask.All);
                            nodes.add(s.nodeID);
                        }

                        if (s.timestampValid) {
                            // calculate orientation
                            if (!s.hasOrientation) {
                                OrientationFilter filter = OrientationFilters.getFilter(serialProtocol.toString(), s.nodeID);
                                if (filter == null) {
                                    filter = new ImuFilter();
                                    OrientationFilters.updateFilter(serialProtocol.toString(), s.nodeID, filter);
                                }
                                if (filter != null) {
                                    double timeDiff = lastTimestamp.containsKey(s.nodeID) ? s.timestamp - lastTimestamp.get(s.nodeID) : 1 / 200;
                                    double[] quat = filter.update(timeDiff, 
                                                                  s.hasValidAccel ? new double[] { s.ax, s.ay, s.az } : null, 
                                                                  s.hasValidCompass ? new double[] { s.cx, s.cy, s.cz } : null,
                                                                  s.hasValidGyro ? new double[] { s.gx, s.gy, s.gz } : null);
                                    s.quat1 = quat[0];
                                    s.quat2 = quat[1];
                                    s.quat3 = quat[2];
                                    s.quat4 = quat[3];
                                    s.hasOrientation = true;
                                }
                                lastTimestamp.put(s.nodeID, s.timestamp);
                            }

                            //System.out.println(s.toCsvLine());
                            if (s.gps != null) {
                                System.out.println(s.gps.toCsvLine());
                            }
                        }
                    }
                }
            }
        } catch (SerialPortException ex) {
            abort = true;
            System.out.println(ex);
            if (crt != null)
                crt.interrupt();
            if (cct != null)
                cct.interrupt();
        } finally {
            try {
                serialPort.closePort();
            } catch (SerialPortException e) {
                System.out.println(e);
            }
        }
    }

    @Override
    public void sendBytes(byte[] data) {
        // Send bytes automatically generated by the library to the serial port.
        try {
            if (serialPort != null && serialPort.isOpened()) {
                serialPort.writeBytes(data);
            }
        } catch (SerialPortException e) {
            e.printStackTrace();
        }
    }

    /**
     * Thread used to automatically request configurations for new nodes.
     */
    class ConfigRequestThread extends Thread {

        private ConcurrentHashMap<Pair<Integer, Mask>, Integer> requestMap = new ConcurrentHashMap<>();
        private static final int NUM_RETRIES = 10;
        private volatile int broadcastRetries = NUM_RETRIES;

        @Override
        public void run() {

            try {

                while (!abort) {
                    if (broadcastRetries > 0) {
                        try {
                            byte[] broadCastReq;
                            Config confReq = new Config(Mask.All);
                            broadCastReq = serialProtocol.createMessage(confReq);
                            serialPort.writeBytes(broadCastReq);
                            broadcastRetries--;
                        } catch (InertiaException e) {
                            System.err.println(e.getMessage());
                        }
                    } else {
                        Iterator<Map.Entry<Pair<Integer, Mask>, Integer>> iter = requestMap.entrySet().iterator();
                        while (iter.hasNext()) {
                            try {
                                Map.Entry<Pair<Integer, Mask>, Integer> entry = iter.next();
                                int nodeID = entry.getKey().first;
                                Mask configMask = entry.getKey().second;
                                Config confReq = new Config(configMask);
                                byte[] nodeReq = serialProtocol.createMessage(confReq, nodeID);
                                serialPort.writeBytes(nodeReq);
                                entry.setValue(entry.getValue() - 1);

                                if (entry.getValue() <= 0)
                                    iter.remove();
                            } catch (InertiaException e) {
                                System.err.println(e.getMessage());
                            }
                        }
                    }

                    try {
                        Thread.sleep(1000);
                    } catch (InterruptedException e) {
                    }
                }
            } catch (SerialPortException e) {
                e.printStackTrace();
            }

        }

        /**
         * Request a configuration for a node.
         *
         * @param nodeID
         *            The node ID
         * @param config
         *            The configuration mask
         */
        public void requestConfig(int nodeID, Mask config) {
            requestMap.put(new Pair<>(nodeID, config), NUM_RETRIES);
        }

        /**
         * Indicate a configuration is received.
         *
         * @param nodeID
         *            the node ID
         */
        public void configReceived(int nodeID) {
            broadcastRetries = 0;

            Iterator<Map.Entry<Pair<Integer, Mask>, Integer>> iter = requestMap.entrySet().iterator();
            while (iter.hasNext()) {
                Map.Entry<Pair<Integer, Mask>, Integer> entry = iter.next();
                if (nodeID == entry.getKey().first) {
                    iter.remove();
                }
            }
        }
    }

    /**
     * Thread used to automatically change configurations.
     */
    class ConfigChangeThread extends Thread {

        @Override
        public void run() {

            try {

                // Toggle transmitting data of all nodes every 10 seconds.
                boolean toggle = false;
                while (!abort) {

                    try {
                        Thread.sleep(10000);
                    } catch (InterruptedException e) {
                        continue;
                    }

                    Set<Integer> nodes = serialProtocol.getTransformerIDs();
                    for (Integer nodeID : nodes) {
                        Transformer tfm = serialProtocol.getTransformer(nodeID);
                        if (tfm.hasGlobalSamplingConfig()) {
                            System.err.println("Changing global config for node " + nodeID + " (transmitting=" + toggle + ")");
                            GlobalSamplingConfig gsc = tfm.getGlobalSamplingConfig();
                            gsc.transmitEnable = toggle;

                            try {
                                Config changedConfig = new Config(gsc);
                                byte[] bytes = serialProtocol.createMessage(changedConfig, nodeID);
                                serialPort.writeBytes(bytes);
                                scheduleRequest(nodeID, gsc.getMask(), 500);
                            } catch (InertiaException e) {
                                System.err.println(e.getMessage());
                            }
                        }
                    }
                    toggle = !toggle;
                }
            } catch (SerialPortException e) {
                e.printStackTrace();
            }

        }

        /**
         * Schedule a new configuration request.
         *
         * @param nodeID
         *            The node ID
         * @param config
         *            The configuration mask
         * @param delayMS
         *            Time in milliseconds before requesting
         */
        public void scheduleRequest(final int nodeID, final Mask config, final int delayMS) {
            new java.util.Timer().schedule(new java.util.TimerTask() {
                @Override
                public void run() {
                    crt.requestConfig(nodeID, config);
                }
            }, delayMS);
        }
    }

    /**
     * Simple Pair
     *
     * @param <A>
     *            First item
     * @param <B>
     *            Second item
     */
    class Pair<A, B> {
        public A first;
        public B second;

        public Pair(A first, B second) {
            super();
            this.first = first;
            this.second = second;
        }
    }
}
